class Resources:
    def __init__(self, cores_min=None, cores_max=None, ram_min=None, ram_max=None):
        self.coresMin = cores_min
        self.coresMax = cores_max
        self.ramMin = ram_min
        self.ramMax = ram_max
