"""

It would be amazing to calculate the Levenshtein distance between two strings, so we could do a:
    "Did you mean: 'thisvariable'"

When you're connecting steps together.

Resources:
    - https://stackoverflow.com/a/5859823/2860731
    - http://stevehanov.ca/blog/index.php?id=114
"""
